import io as _io
import os as _os


def embeddings(folder: str, text_vectorizer, embedding_weights = None, embedding_layer = None, model = None, embedding_layer_name = None) -> None:
    """
    Exports the embedding weights and the vocabulary of the text vectorizer.

    Use in conjunction with:

    https://projector.tensorflow.org

    :param folder: The target folder to save the weights and vocabulary to.
    :param text_vectorizer: A TextVectorization layer
    :param embedding_layer: An Embedding layer
    """
    # Check if target directory exists, if not, create it
    _os.makedirs(folder, exist_ok=True)

    out_v = _io.open(_os.path.join(folder, 'vectors.tsv'), 'w', encoding='utf-8')
    out_m = _io.open(_os.path.join(folder, 'metadata.tsv'), 'w', encoding='utf-8')

    if embedding_weights is None and embedding_layer is not None:
        embedding_weights = embedding_layer.get_weights()[0]
    elif embedding_weights is None and embedding_layer_name is not None and model is not None:
        embedding_weights = model.get_layer(name=embedding_layer_name).get_weights()[0]

    if embedding_weights is None:
        raise 'No embedding weights specified, either specify embedding_weights or embedding_layer or model and embedding_layer_name.'

    vocabulary = text_vectorizer.get_vocabulary()

    for index, word in enumerate(vocabulary):
        if index == 0:
            continue  # skip 0, it's padding.
        vec = embedding_weights[index]
        out_v.write('\t'.join([str(x) for x in vec]) + "\n")
        out_m.write(word + "\n")
    out_v.close()
    out_m.close()



import tensorflow as _tf
import so_ml_tools as _soml


def describe_pipeline(dataset: _tf.data.Dataset):
    """
    Describes the different steps in the `tf.data.Dataset` pipeline.

    :param dataset: the dataset.
    """
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    print(f"{input_dataset.__class__.__name__}", end="")

    while hasattr(input_dataset, '_input_dataset'):
        print(" -> ", end="")
        input_dataset = input_dataset._input_dataset
        print(f"{input_dataset.__class__.__name__}", end="")


def describe_inputs_and_outputs(dataset: _tf.data.Dataset) -> None:
    """
    Describes the input / output shapes and dtype's for the given `tf.data.Dataset`

    Args:
        dataset: a `tf.data.Dataset`
    """
    if isinstance(dataset.element_spec, tuple):
        inputs = dataset.element_spec[0]
        outputs = dataset.element_spec[1]

        print(f"          | {'dtype':<16} | shape")
        # In case of a single input
        if isinstance(inputs, _tf.TensorSpec):
            print(f"Input     | {str(inputs.dtype.name):<16} | {str(inputs.shape)}")
        else:
            # In case of multiple concatenated datasets.
            for i, input in enumerate(inputs):
                print(f"Input  #{i} | {str(input.dtype.name):<16} | {str(input.shape)}")

        if isinstance(outputs, _tf.TensorSpec):
            print(f"Output    | {str(outputs.dtype.name):<16} | {str(outputs.shape)}")


def add_rescaling_mapping(dataset: _tf.data.Dataset) -> _tf.data.Dataset:
    """
    Adds a rescale mapping to the dataset (0-255 -> 0-1)

    Preferably do this before using `optimize_pipeline`

    Args:
        dataset: a `tf.data.Dataset`

    Returns:
        A `tf.data.Dataset` with rescaling.
    """
    return dataset.map(_rescale)


def _rescale(x, y):
    return x / 255., y


def add_batching(dataset: _tf.data.Dataset, batch_size=32) -> _tf.data.Dataset:
    if is_batched(dataset=dataset):
        print('WARN: Dataset is already batched.')
    return dataset.batch(batch_size=batch_size)


def add_caching(dataset: _tf.data.Dataset) -> _tf.data.Dataset:
    if is_cached(dataset=dataset):
        print('WARN: Dataset is already cached.')
    return dataset.cache()


def add_shuffling(dataset: _tf.data.Dataset, buffer_size=_tf.data.AUTOTUNE) -> _tf.data.Dataset:
    if is_shuffled(dataset=dataset):
        print('WARN: Dataset is already shuffled.')

    if buffer_size == _tf.data.AUTOTUNE:
        if is_batched(dataset=dataset):
            batch_size = get_batch_size(dataset)
            buffer_size = batch_size * 8
        else:
            buffer_size = 1000

        print(f'Buffer size for shuffle has been set to {buffer_size}')
    return dataset.shuffle(buffer_size=buffer_size)


def add_prefetching(dataset: _tf.data.Dataset, buffer_size=_tf.data.AUTOTUNE) -> _tf.data.Dataset:
    if is_prefetched(dataset=dataset):
        print('WARN: Dataset is already prefetched.')
    return dataset.prefetch(buffer_size=buffer_size)


def get_class_names_from_dataset_info(ds_info: dict):
    """
    Returns the labels from the dataset info object which is created from loading a
    TensorFlow Dataset (tensorflow_datasets)

    :param ds_info: The dataset info object
    :return: The labels / class names
    """
    return ds_info["features"]["label"].names


def get_class_names(dataset: _tf.data.Dataset):
    """
    Returns the class names from the dataset
    :param dataset: the dataset from which we want the class names.
    :return: the class names
    """
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not hasattr(input_dataset, 'class_names') and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    if not hasattr(input_dataset, 'class_names'):
        raise TypeError("dataset does not have a ´class_names´ attribute defined.")

    return input_dataset.class_names


def get_labels(dataset: _tf.data.Dataset):
    """
    Returns the labels from a (batched)Dataset

    :param dataset: the dataset from which we want the labels.
    :return: the labels
    """
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not hasattr(input_dataset, '_batch_size') and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    if hasattr(input_dataset, '_batch_size'):
        dataset = dataset.unbatch()

    y_labels = []
    for _, labels in dataset:
        y_labels.append(labels.numpy())

    return y_labels


def get_batch_dataset(dataset: _tf.data.Dataset) -> _tf.data.Dataset:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == '_BatchDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    if input_dataset.__class__.__name__ == '_BatchDataset':
        return input_dataset


def get_batch_size(dataset: _tf.data.Dataset) -> int:
    batch_dataset = get_batch_dataset(dataset=dataset)
    if not batch_dataset:
        # noinspection PyUnresolvedReferences
        # noinspection PyProtectedMember
        return batch_dataset._batch_size

    return -1


def is_batched(dataset: _tf.data.Dataset) -> bool:
    return not get_batch_dataset(dataset=dataset)


def is_prefetched(dataset: _tf.data.Dataset) -> bool:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == '_PrefetchDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    return input_dataset.__class__.__name__ == '_PrefetchDataset'


def is_cached(dataset: _tf.data.Dataset) -> bool:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == 'CacheDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    return input_dataset.__class__.__name__ == 'CacheDataset'


def is_shuffled(dataset: _tf.data.Dataset) -> bool:
    if not isinstance(dataset, _tf.data.Dataset):
        raise TypeError('dataset is not a tf.data.Dataset')

    input_dataset = dataset
    while not input_dataset.__class__.__name__ == '_ShuffleDataset' and hasattr(input_dataset, '_input_dataset'):
        # noinspection PyProtectedMember
        input_dataset = input_dataset._input_dataset

    return input_dataset.__class__.__name__ == '_ShuffleDataset'


def show_images_from_dataset(dataset: _tf.data.Dataset, shape=(4, 8)):
    _soml.data.image.show_images_from_dataset(dataset=dataset, shape=shape)



import sklearn as _sklearn
import so_ml_tools as _soml


class FillBackwardImputer(_sklearn.base.BaseEstimator, _sklearn.base.TransformerMixin):

    def __init__(self, add_indicator=False):
        self.add_indicator = add_indicator
        self.column_names_ = None

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        self.column_names_ = list(X.columns)
        _soml.pd.dataframe.fill_nan_with_next_value(dataframe=X, column_names=self.column_names_, inplace=True,
                                                    add_indicator=self.add_indicator)
        return X

    def get_feature_names_out(self, input_features=None):
        if self.add_indicator:
            feature_names = []
            for item in self.column_names_:
                feature_names.append(item)
                feature_names.append(str(item) + '_nan')
        else:
            feature_names = self.column_names_

        return feature_names


class FillForwardImputer(_sklearn.base.BaseEstimator, _sklearn.base.TransformerMixin):

    def __init__(self, add_indicator=False):
        self.add_indicator = add_indicator
        self.column_names_ = None

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        self.column_names_ = list(X.columns)
        _soml.pd.dataframe.fill_nan_with_previous_value(dataframe=X, column_names=self.column_names_, inplace=True,
                                                        add_indicator=self.add_indicator)
        return X

    def get_feature_names_out(self, input_features=None):
        if self.add_indicator:
            feature_names = []
            for item in self.column_names_:
                feature_names.append(item)
                feature_names.append(str(item) + '_nan')
        else:
            feature_names = self.column_names_

        return feature_names


class SimpleGroupByImputer(_sklearn.base.BaseEstimator, _sklearn.base.TransformerMixin):

    def __init__(self, column_name, group_by_column_name, agg_func, remainder_agg_func):
        self.column_name = column_name
        self.group_by_column_name = group_by_column_name
        self.agg_func = agg_func
        self.remainder_agg_func = remainder_agg_func
        self.mean_by_group_ = None
        self.global_mean_ = None

    def fit(self, X, y=None):
        assert self.group_by_column_name in X, (f"fill_nan_with_func_grouped_by: Column '{self.group_by_column_name}' "
                                                f"does not exist in dataframe, did you add "
                                                f"'{self.group_by_column_name}' to the list of columns?")
        self.mean_by_group_ = X.groupby(self.group_by_column_name)[self.column_name].agg(self.agg_func)
        self.global_mean_ = X[self.column_name].apply(self.remainder_agg_func)
        return self

    def transform(self, X):
        # Fill in all NaN values where we have a calculated mean value from the group by.
        X[self.column_name].fillna(X[self.group_by_column_name].map(self.mean_by_group_), inplace=True)

        if self.remainder_agg_func:
            # Fill in all the NaN values with the global mean value.
            X[self.column_name].fillna(value=self.global_mean_, inplace=True)

        # Drop the group by column, otherwise it will be part of the feature set.
        X.drop(self.group_by_column_name, axis=1, inplace=True)

        return X

    def get_feature_names_out(self, input_features=None):
        return [self.column_name]


from setuptools import setup, find_packages
from so_ml_tools import __version__

DESCRIPTION = 'Machine Learning Convenience Functions'

with open("requirements.txt", "r") as fh:
    install_requires = fh.read()

# Setting up
setup(
    # the name must match the folder name 'verysimplemodule'
    name="so-ml-tools",
    version=__version__,
    author="Ronald Mathies",
    description="Utilities library for TensorFlow",
    long_description_content_type="text/plain",
    long_description="A various set of utility functions that can be used in conjunction with Tensorflow.",
    packages=find_packages(),
    install_requires=install_requires,
    keywords=['python', 'machine', 'deep', 'learning', 'tensorflow'],
    classifiers=[
        "Programming Language :: Python :: 3"
    ]
)


from . import (
    dataset,
    tensor,
    model,
    fit,
    mixed_precision,
    tensorboard
)

__all__ = [
    "dataset",
    "tensor",
    "model",
    "fit",
    "mixed_precision",
    "tensorboard"
]
